﻿; PB 5.73 LTS
;
; Rules
; =====
; After an election, several parties have obtained a number of seats in parliament.
; The government needs the absolute majority of seats, i.e. at least 1 seat more
; than all other parties combined. For reasons of political stability, it might be
; desirable for the government to have a larger majority. When calling the
; GetCoalitions() function, choose an appropriate value for the 'majority' parameter.
; The goal of this program is to make proposals for a government that consists of the
; smallest possible number of parties. If e.g. {A,B} is a solution, {A,B,C} will not
; be taken into account. The proposals can be used to decide which coalitions would
; make political sense.
;
; Note: The party with the most seats does not have to be involved in government
;       (this is the law in Germany, for example).

XIncludeFile "set.pbi"


DeclareModule Coa
   Structure Coa Extends Set::Set
      Rest.i
   EndStructure
   
   Declare AssignSeats   (*parliament.Set::Set, party$, numSeats.i)
   Declare GetCoalitions (*parliament.Set::Set, List coalition.Coa(), ignore$="{}", majority.i=1)
   
   ; The following procedure is optional and only for convenience.
   ; You can do other things with the list of generated coalitions.
   Declare ShowCoalitions (title$, List coalition.Coa(), width.i)
EndDeclareModule


Module Coa
   EnableExplicit
   
   NewMap s_Seats.i()
   Define s_TotalSeats.i
   
   
   Procedure AssignSeats (*parliament.Set::Set, party$, numSeats.i)
      ; in : *parliament: set of all parties in the parliament
      ;      party$     : name of a party in the parliament
      ;      numSeats   : number of seats of that party
      ; out: *parliament: updated set of all parties in the parliament
      Shared s_Seats(), s_TotalSeats
      
      Set::AddElement(*parliament, party$)
      s_Seats(party$) = numSeats
      s_TotalSeats + numSeats
   EndProcedure
   
   
   Procedure.i _NumSeats (*parties.Set::Set)
      ; in : *parties: set of parties
      ; out: return value: sum of seats of all parties in the set
      Shared s_Seats()
      Protected ret.i=0
      
      ForEach *parties\Element()
         ret + s_Seats(MapKey(*parties\Element()))
      Next
      
      ProcedureReturn ret
   EndProcedure
   
   
   Procedure.i _GetCoalitionsK (*parliament.Set::Set, requiredSeats.i, parties.i, List coalition.Coa())
      ; in : *parliament  : set of all parties in the parliament
      ;      requiredSeats: number of required seats for the desired majority
      ;      parties      : number of parties in the current coalition
      ;      coalition()  : list of coalitions that have been found so far by previous calls to this procedure
      ; out: coalition(): list of coalitions, possibly updated
      Shared s_Seats(), s_TotalSeats
      Protected.i nxt, coaSeats, newCoa
      Protected curCoalition.Coa
      Protected NewList coalitionK.Coa()
      
      nxt = Set::FirstSubsetK(*parliament, curCoalition, parties)
      While nxt <> -1
         coaSeats = _NumSeats(curCoalition)
         If coaSeats >= requiredSeats
            newCoa = #True
            ForEach coalition()
               If Set::IsSubset(curCoalition, coalition())
                  ; if 'curCoalition' would only be an extension of a coalition that was already found
                  newCoa = #False
                  Break
               EndIf
            Next
            
            If newCoa = #True
               ForEach curCoalition\Element()
                  curCoalition\Element() = s_Seats(MapKey(curCoalition\Element()))
               Next   
               curCoalition\Value = coaSeats
               curCoalition\Rest  = s_TotalSeats - coaSeats
               AddElement(coalitionK())
               CopyStructure(@curCoalition, @coalitionK(), Coa)
            EndIf
         EndIf
         
         nxt = Set::NextSubsetK(*parliament, curCoalition)
      Wend
      
      SortStructuredList(coalitionK(), #PB_Sort_Descending, OffsetOf(Coa\Value), TypeOf(Coa\Value))
      MergeLists(coalitionK(), coalition())
   EndProcedure
   
   
   Procedure GetCoalitions (*parliament.Set::Set, List coalition.Coa(), ignore$="{}", majority.i=1)
      ; in : *parliament: set of all parties in the parliament
      ;      ignore$    : set of parties that are ignored because they are politically unacceptable for a coalition
      ;      majority   : >= 1  (The coalition needs at least 1 seat more than all other parties in the parliament.
      ;                          For the sake of political stability, this value can be increased.)
      ; out: coalition(): list containing 1 party or all possible coalitions,
      ;                   sorted ascending by number of parties and then descending by number of seats
      Shared s_Seats(), s_TotalSeats
      Protected.i requiredSeats, parties
      Protected skip.Set::Set
      
      ClearList(coalition())
      requiredSeats = Int(s_TotalSeats / 2) + majority
      
      Set::FromString(skip, ignore$)
      Set::Difference(*parliament, skip, *parliament)
      
      For parties = 1 To *parliament\NumElements - 1
         _GetCoalitionsK(*parliament, requiredSeats, parties, coalition())
      Next
      
      ; Reset shared variables
      ClearMap(s_Seats())
      s_TotalSeats = 0
   EndProcedure
   
   ;------------------------------------------------------------------------
   
   Structure Party
      Name$
      Seats.i
   EndStructure
   
   Procedure.s _OneCoalition (*a.Coa, width.i, maxDigits.i)
      ; in : *a       : set that represents one coalition
      ;      width    : minimum width for displaying the party names
      ;      maxDigits: maximum number of digits needed for displaying number of seats
      ; out: return value: all elements in the given set, their combined number of seats,
      ;                    and the remaining number of seats in the parliament
      Protected ret$
      Protected NewList gov.Party()
      
      ; Copy the parties in the set to a list
      ForEach *a\Element()
         AddElement(gov())
         gov()\Name$ = MapKey(*a\Element())
         gov()\Seats = *a\Element()
      Next
      
      ; Sort list of parties according to number of seats
      SortStructuredList(gov(), #PB_Sort_Descending, OffsetOf(Party\Seats), TypeOf(Party\Seats))
      
      ; Write names of parties to a string
      FirstElement(gov())
      ret$ = "|" + gov()\Name$ + "|"
      While NextElement(gov())
         ret$ + " + |" + gov()\Name$ + "|"
      Wend
      
      If Len(ret$) < width
         ret$ = LSet(ret$, width, " ")
      EndIf
      
      ProcedureReturn ret$ + " = " + RSet(Str(*a\Value), maxDigits) + 
                             "  (vs. " + RSet(Str(*a\Rest), maxDigits) + ")"
   EndProcedure
   
   
   Procedure ShowCoalitions (title$, List coalition.Coa(), width.i)
      ; -- Show all generated coalitions,
      ;    sorted ascending by number of parties and then descending by number of seats.
      ;    In each coalition, the parties are sorted descending by number of seats.
      ; in: title$     : title
      ;     coalition(): list containing 1 party or all generated coalitions
      ;     width      : minimum width for displaying the party names
      Static coaParties.i
      Protected maxDigits.i
      
      If FirstElement(coalition())
         Debug title$
         If ListSize(coalition()) = 1
            Debug "One party has the absolute majority:"
         Else
            Debug "There are " + ListSize(coalition()) + " possible coalitions:"
         EndIf
         Debug ""
         
         maxDigits = Round(Log(coalition()\Value + coalition()\Rest) / Log(10), #PB_Round_Down) + 1
         Repeat
            If coaParties <> 0 And coaParties <> Set::Size(coalition())
               Debug ""
            EndIf
            coaParties = Set::Size(coalition())
            
            Debug _OneCoalition(coalition(), width, maxDigits)
         Until NextElement(coalition()) = #False
      EndIf
      
      coaParties = 0    ; reset static variable
   EndProcedure
EndModule


If #PB_Compiler_Debugger = #False
   MessageRequester("set_demo_coalitions", "Switch Debugger on!")
   End
EndIf

EnableExplicit

Define parliament.Set::Set
NewList coalition.Coa::Coa()

; ------------------------------------------------
; -- German Bundestag Election 2013
Coa::AssignSeats(parliament, "CDU/CSU", 311)
Coa::AssignSeats(parliament, "SPD"    , 193)
Coa::AssignSeats(parliament, "LINKE"  ,  64)
Coa::AssignSeats(parliament, "GRÜNE"  ,  63)
; ------------------------------------------------

Coa::GetCoalitions(parliament, coalition())
Coa::ShowCoalitions("-- Bundestag 2013", coalition(), 34)

Debug ""
Debug "===================================================="
Debug ""

Set::Clear(parliament)

; ------------------------------------------------
; -- German Bundestag Election 2017
Coa::AssignSeats(parliament, "CDU/CSU", 246)
Coa::AssignSeats(parliament, "SPD"    , 153)
Coa::AssignSeats(parliament, "AfD"    ,  94)
Coa::AssignSeats(parliament, "FDP"    ,  80)
Coa::AssignSeats(parliament, "LINKE"  ,  69)
Coa::AssignSeats(parliament, "GRÜNE"  ,  67)
; ------------------------------------------------

Coa::GetCoalitions(parliament, coalition(), "{AfD}")
Coa::ShowCoalitions("-- Bundestag 2017", coalition(), 34)

; IDE Options = PureBasic 5.73 LTS (Windows - x64)
; Folding = --
; EnableXP
; EnableUser
; DPIAware
; EnableExeConstant